@@ -1,6 +1,6 @@
 // SPDX-License-Identifier: MPL-2.0
 
-use crate::signaller::{Signallable, SignallableImpl};
+use crate::signaller::{Signallable, SignallableExt, SignallableImpl};
 use crate::utils::{
     build_link_header, build_reqwest_client, parse_redirect_location, set_ice_servers, wait,
     wait_async, WaitError,
@@ -12,15 +12,15 @@ use gst::prelude::*;
 use gst::subclass::prelude::*;
 use gst_sdp::SDPMessage;
 use gst_webrtc::{WebRTCICEGatheringState, WebRTCSessionDescription};
-use once_cell::sync::Lazy;
 use reqwest::header::HeaderMap;
 use reqwest::header::HeaderValue;
 use reqwest::StatusCode;
+use std::collections::HashMap;
+use std::sync::LazyLock;
 use std::sync::Mutex;
 
-use core::time::Duration;
-use crossbeam_channel::unbounded;
 use std::net::SocketAddr;
+use tokio::sync::mpsc;
 use url::Url;
 use warp::{
     http,
@@ -31,7 +31,7 @@ use warp::{
     Filter, Reply,
 };
 
-static CAT: Lazy<gst::DebugCategory> = Lazy::new(|| {
+static CAT: LazyLock<gst::DebugCategory> = LazyLock::new(|| {
     gst::DebugCategory::new(
         "webrtc-whip-signaller",
         gst::DebugColorFlags::empty(),
@@ -47,7 +47,6 @@ const ENDPOINT_PATH: &str = "endpoint";
 const RESOURCE_PATH: &str = "resource";
 const DEFAULT_HOST_ADDR: &str = "http://127.0.0.1:8080";
 const DEFAULT_STUN_SERVER: Option<&str> = Some("stun://stun.l.google.com:19303");
-const DEFAULT_PRODUCER_PEER_ID: Option<&str> = Some("whip-client");
 const CONTENT_SDP: &str = "application/sdp";
 const CONTENT_TRICKLE_ICE: &str = "application/trickle-ice-sdpfrag";
 
@@ -99,7 +98,7 @@ impl WhipClient {
     fn handle_future_error(&self, err: WaitError) {
         match err {
             WaitError::FutureAborted => {
-                gst::warning!(CAT, imp: self, "Future aborted")
+                gst::warning!(CAT, imp = self, "Future aborted")
             }
             WaitError::FutureError(err) => self.raise_error(err.to_string()),
         };
@@ -120,12 +119,12 @@ impl WhipClient {
                 self.raise_error("Local description is not set".to_string());
                 return;
             }
-            Some(offer) => offer,
+            Some(offer) => self.obj().munge_sdp("unique", &offer),
         };
 
         gst::debug!(
             CAT,
-            imp: self,
+            imp = self,
             "Sending offer SDP: {:?}",
             offer_sdp.sdp().as_text()
         );
@@ -189,11 +188,11 @@ impl WhipClient {
         let sdp = offer.sdp();
         let body = sdp.as_text().unwrap();
 
-        gst::debug!(CAT, imp: self, "Using endpoint {}", endpoint.as_str());
+        gst::debug!(CAT, imp = self, "Using endpoint {}", endpoint.as_str());
         let mut headermap = HeaderMap::new();
         headermap.insert(
             reqwest::header::CONTENT_TYPE,
-            HeaderValue::from_static("application/sdp"),
+            HeaderValue::from_static(CONTENT_SDP),
         );
 
         if let Some(token) = auth_token.as_ref() {
@@ -228,7 +227,7 @@ impl WhipClient {
         redirects: u8,
         webrtcbin: &gst::Element,
     ) {
-        gst::debug!(CAT, imp: self, "Parsing endpoint response");
+        gst::debug!(CAT, imp = self, "Parsing endpoint response");
 
         let endpoint;
         let use_link_headers;
@@ -279,7 +278,7 @@ impl WhipClient {
 
                 let url = reqwest::Url::parse(endpoint.as_str()).unwrap();
 
-                gst::debug!(CAT, imp: self, "WHIP resource: {:?}", location);
+                gst::debug!(CAT, imp = self, "WHIP resource: {:?}", location);
 
                 let url = match url.join(location) {
                     Ok(joined_url) => joined_url,
@@ -354,7 +353,7 @@ impl WhipClient {
 
                             gst::debug!(
                                 CAT,
-                                imp: self,
+                                imp = self,
                                 "Redirecting endpoint to {}",
                                 redirect_url.as_str()
                             );
@@ -409,7 +408,7 @@ impl WhipClient {
             );
         }
 
-        gst::debug!(CAT, imp: self, "DELETE request on {}", resource_url);
+        gst::debug!(CAT, imp = self, "DELETE request on {}", resource_url);
         let client = build_reqwest_client(reqwest::redirect::Policy::default());
         let future = async {
             client
@@ -428,14 +427,14 @@ impl WhipClient {
         let res = wait(&self.canceller, future, timeout);
         match res {
             Ok(r) => {
-                gst::debug!(CAT, imp: self, "Response to DELETE : {}", r.status());
+                gst::debug!(CAT, imp = self, "Response to DELETE : {}", r.status());
             }
             Err(e) => match e {
                 WaitError::FutureAborted => {
-                    gst::warning!(CAT, imp: self, "DELETE request aborted")
+                    gst::warning!(CAT, imp = self, "DELETE request aborted")
                 }
                 WaitError::FutureError(e) => {
-                    gst::error!(CAT, imp: self, "Error on DELETE request : {}", e)
+                    gst::error!(CAT, imp = self, "Error on DELETE request : {}", e)
                 }
             },
         };
@@ -455,32 +454,34 @@ impl SignallableImpl for WhipClient {
             glib::closure!(|signaller: &super::WhipClientSignaller,
                             _consumer_identifier: &str,
                             webrtcbin: &gst::Element| {
-                let obj_weak = signaller.downgrade();
-                webrtcbin.connect_notify(Some("ice-gathering-state"), move |webrtcbin, _pspec| {
-                    let Some(obj) = obj_weak.upgrade() else {
-                        return;
-                    };
+                webrtcbin.connect_notify(
+                    Some("ice-gathering-state"),
+                    glib::clone!(
+                        #[weak]
+                        signaller,
+                        move |webrtcbin, _pspec| {
+                            let state = webrtcbin
+                                .property::<WebRTCICEGatheringState>("ice-gathering-state");
 
-                    let state =
-                        webrtcbin.property::<WebRTCICEGatheringState>("ice-gathering-state");
+                            match state {
+                                WebRTCICEGatheringState::Gathering => {
+                                    gst::info!(CAT, obj = signaller, "ICE gathering started");
+                                }
+                                WebRTCICEGatheringState::Complete => {
+                                    gst::info!(CAT, obj = signaller, "ICE gathering complete");
 
-                    match state {
-                        WebRTCICEGatheringState::Gathering => {
-                            gst::info!(CAT, obj: obj, "ICE gathering started");
-                        }
-                        WebRTCICEGatheringState::Complete => {
-                            gst::info!(CAT, obj: obj, "ICE gathering complete");
+                                    let webrtcbin = webrtcbin.clone();
 
-                            let webrtcbin = webrtcbin.clone();
-
-                            RUNTIME.spawn(async move {
+                                    RUNTIME.spawn(async move {
                                 /* Note that we check for a valid WHIP endpoint in change_state */
-                                obj.imp().send_offer(&webrtcbin).await
+                                signaller.imp().send_offer(&webrtcbin).await
                             });
+                                }
+                                _ => (),
+                            }
                         }
-                        _ => (),
-                    }
-                });
+                    ),
+                );
             }),
         );
 
@@ -534,8 +535,15 @@ impl ObjectSubclass for WhipClient {
 
 impl ObjectImpl for WhipClient {
     fn properties() -> &'static [glib::ParamSpec] {
-        static PROPERTIES: Lazy<Vec<glib::ParamSpec>> = Lazy::new(|| {
-            vec![glib::ParamSpecString::builder("whip-endpoint")
+        static PROPERTIES: LazyLock<Vec<glib::ParamSpec>> = LazyLock::new(|| {
+            vec![
+                glib::ParamSpecBoolean::builder("manual-sdp-munging")
+                    .nick("Manual SDP munging")
+                    .blurb("Whether the signaller manages SDP munging itself")
+                    .default_value(false)
+                    .read_only()
+                    .build(),
+                glib::ParamSpecString::builder("whip-endpoint")
                     .nick("WHIP Endpoint")
                     .blurb("The WHIP server endpoint to POST SDP offer to.
                         e.g.: https://example.com/whip/endpoint/room1234")
@@ -592,6 +600,7 @@ impl ObjectImpl for WhipClient {
 
     fn property(&self, _id: usize, pspec: &glib::ParamSpec) -> glib::Value {
         match pspec.name() {
+            "manual-sdp-munging" => true.to_value(),
             "whip-endpoint" => {
                 let settings = self.settings.lock().unwrap();
                 settings.whip_endpoint.to_value()
@@ -616,27 +625,14 @@ impl ObjectImpl for WhipClient {
 // WHIP server implementation
 
 #[derive(Debug)]
-enum WhipServerState {
-    Idle,
-    Negotiating,
-    Ready,
-}
-
-impl Default for WhipServerState {
-    fn default() -> Self {
-        Self::Idle
-    }
-}
-
 struct WhipServerSettings {
     stun_server: Option<String>,
     turn_servers: gst::Array,
     host_addr: Url,
-    producer_peer_id: Option<String>,
     timeout: u32,
     shutdown_signal: Option<tokio::sync::oneshot::Sender<()>>,
     server_handle: Option<tokio::task::JoinHandle<()>>,
-    sdp_answer: Option<crossbeam_channel::Sender<Option<SDPMessage>>>,
+    sdp_answer: HashMap<String, mpsc::Sender<Option<SDPMessage>>>,
 }
 
 impl Default for WhipServerSettings {
@@ -645,65 +641,80 @@ impl Default for WhipServerSettings {
             host_addr: Url::parse(DEFAULT_HOST_ADDR).unwrap(),
             stun_server: DEFAULT_STUN_SERVER.map(String::from),
             turn_servers: gst::Array::new(Vec::new() as Vec<glib::SendValue>),
-            producer_peer_id: DEFAULT_PRODUCER_PEER_ID.map(String::from),
             timeout: DEFAULT_TIMEOUT,
             shutdown_signal: None,
             server_handle: None,
-            sdp_answer: None,
+            sdp_answer: HashMap::new(),
         }
     }
 }
 
+#[derive(Default)]
 pub struct WhipServer {
-    state: Mutex<WhipServerState>,
     settings: Mutex<WhipServerSettings>,
 }
 
-impl Default for WhipServer {
-    fn default() -> Self {
-        Self {
-            settings: Mutex::new(WhipServerSettings::default()),
-            state: Mutex::new(WhipServerState::default()),
-        }
-    }
-}
-
 impl WhipServer {
     pub fn on_webrtcbin_ready(&self) -> RustClosure {
         glib::closure!(|signaller: &super::WhipServerSignaller,
-                        _producer_identifier: &str,
+                        session_id: &str,
                         webrtcbin: &gst::Element| {
-            let obj_weak = signaller.downgrade();
-            webrtcbin.connect_notify(Some("ice-gathering-state"), move |webrtcbin, _pspec| {
-                let obj = match obj_weak.upgrade() {
-                    Some(obj) => obj,
-                    None => return,
-                };
+            webrtcbin.connect_notify(
+                Some("ice-gathering-state"),
+                glib::clone!(
+                    #[weak]
+                    signaller,
+                    #[to_owned]
+                    session_id,
+                    move |webrtcbin, _pspec| {
+                        let state =
+                            webrtcbin.property::<WebRTCICEGatheringState>("ice-gathering-state");
 
-                let state = webrtcbin.property::<WebRTCICEGatheringState>("ice-gathering-state");
+                        match state {
+                            WebRTCICEGatheringState::Gathering => {
+                                gst::info!(CAT, obj = signaller, "ICE gathering started");
+                            }
+                            WebRTCICEGatheringState::Complete => {
+                                gst::info!(
+                                    CAT,
+                                    obj = signaller,
+                                    "ICE gathering complete for {session_id}"
+                                );
+                                let ans: Option<gst_sdp::SDPMessage>;
+                                let mut settings = signaller.imp().settings.lock().unwrap();
+                                if let Some(answer_desc) = webrtcbin
+                                    .property::<Option<WebRTCSessionDescription>>(
+                                        "local-description",
+                                    )
+                                {
+                                    ans = Some(answer_desc.sdp().to_owned());
+                                } else {
+                                    ans = None;
+                                }
+                                let tx = settings
+                                    .sdp_answer
+                                    .remove(&session_id)
+                                    .expect("SDP answer Sender needs to be valid");
 
-                match state {
-                    WebRTCICEGatheringState::Gathering => {
-                        gst::info!(CAT, obj: obj, "ICE gathering started");
-                    }
-                    WebRTCICEGatheringState::Complete => {
-                        gst::info!(CAT, obj: obj, "ICE gathering complete");
-                        let ans: Option<gst_sdp::SDPMessage>;
-                        let settings = obj.imp().settings.lock().unwrap();
-                        if let Some(answer_sdp) = webrtcbin
-                            .property::<Option<WebRTCSessionDescription>>("local-description")
-                        {
-                            ans = Some(answer_sdp.sdp());
-                        } else {
-                            ans = None;
-                        }
-                        if let Some(tx) = &settings.sdp_answer {
-                            tx.send(ans).unwrap()
+                                RUNTIME.spawn(glib::clone!(
+                                    #[strong]
+                                    signaller,
+                                    async move {
+                                        if let Err(e) = tx.send(ans).await {
+                                            gst::error!(
+                                                CAT,
+                                                obj = signaller,
+                                                "Failed to send SDP {e}"
+                                            );
+                                        }
+                                    }
+                                ));
+                            }
+                            _ => (),
                         }
                     }
-                    _ => (),
-                }
-            });
+                ),
+            );
         })
     }
 
@@ -717,57 +728,24 @@ impl WhipServer {
         //FIXME: add state checking once ICE trickle is implemented
     }
 
-    async fn delete_handler(&self, _id: String) -> Result<impl warp::Reply, warp::Rejection> {
-        let mut state = self.state.lock().unwrap();
-        match *state {
-            WhipServerState::Ready => {
-                // FIXME: session-ended will make webrtcsrc send EOS
-                // and producer-removed is not handled
-                // Need to address the usecase where when the client terminates
-                // the webrtcsrc should be running without sending EOS and reset
-                // for next client connection like a usual server
-
-                self.obj().emit_by_name::<bool>("session-ended", &[&ROOT]);
-
-                gst::info!(CAT, imp:self, "Ending session");
-                *state = WhipServerState::Idle;
-                Ok(warp::reply::reply().into_response())
-            }
-            _ => {
-                gst::error!(CAT, imp: self, "DELETE requested in {state:?} state. Can't Proceed");
-                let res = http::Response::builder()
-                    .status(http::StatusCode::CONFLICT)
-                    .body(Body::from(String::from("Session not Ready")))
-                    .unwrap();
-                Ok(res)
-            }
+    async fn delete_handler(&self, id: String) -> Result<impl warp::Reply, warp::Rejection> {
+        if self
+            .obj()
+            .emit_by_name::<bool>("session-ended", &[&id.as_str()])
+        {
+            gst::info!(CAT, imp = self, "Ended session {id}");
+        } else {
+            gst::info!(CAT, imp = self, "Failed to End session {id}");
+            // FIXME: Do we send a different response
         }
+        Ok(warp::reply::reply().into_response())
     }
 
+    #[allow(clippy::single_match)]
     async fn options_handler(&self) -> Result<impl warp::Reply, warp::Rejection> {
         let settings = self.settings.lock().unwrap();
-        let peer_id = settings.producer_peer_id.clone().unwrap();
         drop(settings);
 
-        let mut state = self.state.lock().unwrap();
-        match *state {
-            WhipServerState::Idle => {
-                self.obj()
-                    .emit_by_name::<()>("session-started", &[&ROOT, &peer_id]);
-                *state = WhipServerState::Negotiating
-            }
-            WhipServerState::Ready => {
-                gst::error!(CAT, imp: self, "OPTIONS requested in {state:?} state. Can't proceed");
-                let res = http::Response::builder()
-                    .status(http::StatusCode::CONFLICT)
-                    .body(Body::from(String::from("Session active already")))
-                    .unwrap();
-                return Ok(res);
-            }
-            _ => {}
-        };
-        drop(state);
-
         let mut links = HeaderMap::new();
         let settings = self.settings.lock().unwrap();
         match &settings.stun_server {
@@ -776,7 +754,7 @@ impl WhipServer {
                     links.append(LINK, HeaderValue::from_str(stun_link.as_str()).unwrap());
                 }
                 Err(e) => {
-                    gst::error!(CAT, imp: self, "Failed to parse {stun:?} : {e:?}");
+                    gst::error!(CAT, imp = self, "Failed to parse {stun:?} : {e:?}");
                 }
             },
             None => {}
@@ -785,23 +763,27 @@ impl WhipServer {
         if !settings.turn_servers.is_empty() {
             for turn_server in settings.turn_servers.iter() {
                 if let Ok(turn) = turn_server.get::<String>() {
-                    gst::debug!(CAT, imp: self, "turn server: {}",turn.as_str());
+                    gst::debug!(CAT, imp = self, "turn server: {}", turn.as_str());
                     match build_link_header(turn.as_str()) {
                         Ok(turn_link) => {
                             links.append(LINK, HeaderValue::from_str(turn_link.as_str()).unwrap());
                         }
                         Err(e) => {
-                            gst::error!(CAT, imp: self, "Failed to parse {turn_server:?} : {e:?}");
+                            gst::error!(CAT, imp = self, "Failed to parse {turn_server:?} : {e:?}");
                         }
                     }
                 } else {
-                    gst::debug!(CAT, imp: self, "Failed to get String value of {turn_server:?}");
+                    gst::debug!(
+                        CAT,
+                        imp = self,
+                        "Failed to get String value of {turn_server:?}"
+                    );
                 }
             }
         }
 
         let mut res = http::Response::builder()
-            .header("Access-Post", "application/sdp")
+            .header("Access-Post", CONTENT_SDP)
             .body(Body::empty())
             .unwrap();
 
@@ -815,31 +797,15 @@ impl WhipServer {
         &self,
         body: warp::hyper::body::Bytes,
     ) -> Result<http::Response<warp::hyper::Body>, warp::Rejection> {
-        let mut settings = self.settings.lock().unwrap();
-        let peer_id = settings.producer_peer_id.clone().unwrap();
-        let wait_timeout = settings.timeout;
-        let (tx, rx) = unbounded::<Option<SDPMessage>>();
-        settings.sdp_answer = Some(tx);
-        drop(settings);
-
-        let mut state = self.state.lock().unwrap();
-        match *state {
-            WhipServerState::Idle => {
-                self.obj()
-                    .emit_by_name::<()>("session-started", &[&ROOT, &peer_id]);
-                *state = WhipServerState::Negotiating
-            }
-            WhipServerState::Ready => {
-                gst::error!(CAT, imp: self, "POST requested in {state:?} state. Can't Proceed");
-                let res = http::Response::builder()
-                    .status(http::StatusCode::CONFLICT)
-                    .body(Body::from(String::from("Session active already")))
-                    .unwrap();
-                return Ok(res);
-            }
-            _ => {}
+        let session_id = uuid::Uuid::new_v4().to_string();
+        let (tx, mut rx) = mpsc::channel::<Option<SDPMessage>>(1);
+        let wait_timeout = {
+            let mut settings = self.settings.lock().unwrap();
+            let wait_timeout = settings.timeout;
+            settings.sdp_answer.insert(session_id.clone(), tx);
+            drop(settings);
+            wait_timeout
         };
-        drop(state);
 
         match gst_sdp::SDPMessage::parse_buffer(body.as_ref()) {
             Ok(offer_sdp) => {
@@ -849,43 +815,61 @@ impl WhipServer {
                 );
 
                 self.obj()
-                    .emit_by_name::<()>("session-description", &[&"unique", &offer]);
+                    .emit_by_name::<()>("session-started", &[&session_id, &session_id]);
+                self.obj()
+                    .emit_by_name::<()>("session-description", &[&session_id, &offer]);
             }
             Err(err) => {
-                gst::error!(CAT, imp: self, "Could not parse offer SDP: {err}");
+                gst::error!(CAT, imp = self, "Could not parse offer SDP: {err}");
                 let reply = warp::reply::reply();
                 let res = warp::reply::with_status(reply, http::StatusCode::NOT_ACCEPTABLE);
                 return Ok(res.into_response());
             }
         }
 
-        // We don't want to wait infinitely for the ice gathering to complete.
-        let answer = match rx.recv_timeout(Duration::from_secs(wait_timeout as u64)) {
-            Ok(a) => a,
-            Err(e) => {
-                let reply = warp::reply::reply();
-                let res;
-                if e.is_timeout() {
-                    res = warp::reply::with_status(reply, http::StatusCode::REQUEST_TIMEOUT);
-                    gst::error!(CAT, imp: self, "Timedout waiting for SDP answer");
-                } else {
-                    res = warp::reply::with_status(reply, http::StatusCode::INTERNAL_SERVER_ERROR);
-                    gst::error!(CAT, imp: self, "Channel got disconnected");
+        // We don't support cancellation, instead we depend on the server shutdown oneshot and wait
+        // timeout here. FIXME: Maybe we can do better?
+        let canceller = Mutex::new(None);
+        let result = wait_async(&canceller, rx.recv(), wait_timeout).await;
+
+        let answer = match result {
+            Ok(ans) => match ans {
+                Some(a) => a,
+                None => {
+                    let err = "Channel closed, can't receive SDP".to_owned();
+                    let res = http::Response::builder()
+                        .status(http::StatusCode::INTERNAL_SERVER_ERROR)
+                        .body(Body::from(err))
+                        .unwrap();
+
+                    return Ok(res);
                 }
-                return Ok(res.into_response());
+            },
+            Err(e) => {
+                let err = match e {
+                    WaitError::FutureAborted => "Aborted".to_owned(),
+                    WaitError::FutureError(err) => err.to_string(),
+                };
+                let res = http::Response::builder()
+                    .status(http::StatusCode::INTERNAL_SERVER_ERROR)
+                    .body(Body::from(err))
+                    .unwrap();
+
+                return Ok(res);
             }
         };
 
         let settings = self.settings.lock().unwrap();
         let mut links = HeaderMap::new();
 
+        #[allow(clippy::single_match)]
         match &settings.stun_server {
             Some(stun) => match build_link_header(stun.as_str()) {
                 Ok(stun_link) => {
                     links.append(LINK, HeaderValue::from_str(stun_link.as_str()).unwrap());
                 }
                 Err(e) => {
-                    gst::error!(CAT, imp: self, "Failed to parse {stun:?} : {e:?}");
+                    gst::error!(CAT, imp = self, "Failed to parse {stun:?} : {e:?}");
                 }
             },
             None => {}
@@ -894,17 +878,25 @@ impl WhipServer {
         if !settings.turn_servers.is_empty() {
             for turn_server in settings.turn_servers.iter() {
                 if let Ok(turn) = turn_server.get::<String>() {
-                    gst::debug!(CAT, imp: self, "turn server: {}",turn.as_str());
+                    gst::debug!(CAT, imp = self, "turn server: {}", turn.as_str());
                     match build_link_header(turn.as_str()) {
                         Ok(turn_link) => {
                             links.append(LINK, HeaderValue::from_str(turn_link.as_str()).unwrap());
                         }
                         Err(e) => {
-                            gst::error!(CAT, imp: self, "Failed to  parse {turn_server:?} : {e:?}");
+                            gst::error!(
+                                CAT,
+                                imp = self,
+                                "Failed to  parse {turn_server:?} : {e:?}"
+                            );
                         }
                     }
                 } else {
-                    gst::error!(CAT, imp: self, "Failed to get String value of {turn_server:?}");
+                    gst::error!(
+                        CAT,
+                        imp = self,
+                        "Failed to get String value of {turn_server:?}"
+                    );
                 }
             }
         }
@@ -917,16 +909,16 @@ impl WhipServer {
             match sdp.as_text() {
                 Ok(text) => {
                     ans_text = Ok(text);
-                    gst::debug!(CAT, imp: self, "{ans_text:?}");
+                    gst::debug!(CAT, imp = self, "{ans_text:?}");
                 }
                 Err(e) => {
                     ans_text = Err(format!("Failed to get SDP answer: {e:?}"));
-                    gst::error!(CAT, imp: self, "{e:?}");
+                    gst::error!(CAT, imp = self, "{e:?}");
                 }
             }
         } else {
             let e = "SDP Answer is empty!".to_string();
-            gst::error!(CAT, imp: self, "{e:?}");
+            gst::error!(CAT, imp = self, "{e:?}");
             ans_text = Err(e);
         }
 
@@ -942,10 +934,10 @@ impl WhipServer {
         drop(settings);
 
         // Got SDP answer, send answer in the response
-        let resource_url = "/".to_owned() + ROOT + "/" + RESOURCE_PATH + "/" + &peer_id;
+        let resource_url = "/".to_owned() + ROOT + "/" + RESOURCE_PATH + "/" + &session_id;
         let mut res = http::Response::builder()
             .status(StatusCode::CREATED)
-            .header(CONTENT_TYPE, "application/sdp")
+            .header(CONTENT_TYPE, CONTENT_SDP)
             .header("location", resource_url)
             .body(Body::from(ans_text.unwrap()))
             .unwrap();
@@ -953,10 +945,6 @@ impl WhipServer {
         let headers = res.headers_mut();
         headers.extend(links);
 
-        let mut state = self.state.lock().unwrap();
-        *state = WhipServerState::Ready;
-        drop(state);
-
         Ok(res)
     }
 
@@ -967,10 +955,10 @@ impl WhipServer {
             Ok(v) => {
                 // pick the first vector item
                 addr = v[0];
-                gst::info!(CAT, imp:self, "using {addr:?} as address");
+                gst::info!(CAT, imp = self, "using {addr:?} as address");
             }
             Err(e) => {
-                gst::error!(CAT, imp:self, "error getting addr from uri  {e:?}");
+                gst::error!(CAT, imp = self, "error getting addr from uri  {e:?}");
                 self.obj()
                     .emit_by_name::<()>("error", &[&format!("Unable to start WHIP Server: {e:?}")]);
                 return None;
@@ -983,37 +971,29 @@ impl WhipServer {
 
         let prefix = warp::path(ROOT);
 
-        let self_weak = self.downgrade();
-
         // POST /endpoint
         let post_filter = warp::post()
             .and(warp::path(ENDPOINT_PATH))
             .and(warp::path::end())
             .and(warp::header::exact(CONTENT_TYPE.as_str(), CONTENT_SDP))
             .and(warp::body::bytes())
-            .and_then(move |body| {
-                let s = self_weak.upgrade();
-                async {
-                    let self_ = s.expect("Need to have the ObjectRef");
-                    self_.post_handler(body).await
-                }
-            });
-
-        let self_weak = self.downgrade();
+            .and_then(glib::clone!(
+                #[weak(rename_to = self_)]
+                self,
+                #[upgrade_or_panic]
+                move |body| async move { self_.post_handler(body).await }
+            ));
 
         // OPTIONS /endpoint
         let options_filter = warp::options()
             .and(warp::path(ENDPOINT_PATH))
             .and(warp::path::end())
-            .and_then(move || {
-                let s = self_weak.upgrade();
-                async {
-                    let self_ = s.expect("Need to have the ObjectRef");
-                    self_.options_handler().await
-                }
-            });
-
-        let self_weak = self.downgrade();
+            .and_then(glib::clone!(
+                #[weak(rename_to = self_)]
+                self,
+                #[upgrade_or_panic]
+                move || async move { self_.options_handler().await }
+            ));
 
         // PATCH /resource/:id
         let patch_filter = warp::patch()
@@ -1024,28 +1004,24 @@ impl WhipServer {
                 CONTENT_TYPE.as_str(),
                 CONTENT_TRICKLE_ICE,
             ))
-            .and_then(move |id| {
-                let s = self_weak.upgrade();
-                async {
-                    let self_ = s.expect("Need to have the ObjectRef");
-                    self_.patch_handler(id).await
-                }
-            });
-
-        let self_weak = self.downgrade();
+            .and_then(glib::clone!(
+                #[weak(rename_to = self_)]
+                self,
+                #[upgrade_or_panic]
+                move |id| async move { self_.patch_handler(id).await }
+            ));
 
         // DELETE /resource/:id
         let delete_filter = warp::delete()
             .and(warp::path(RESOURCE_PATH))
             .and(warp::path::param::<String>())
             .and(warp::path::end())
-            .and_then(move |id| {
-                let s = self_weak.upgrade();
-                async {
-                    let self_ = s.expect("Need to have the ObjectRef");
-                    self_.delete_handler(id).await
-                }
-            });
+            .and_then(glib::clone!(
+                #[weak(rename_to = self_)]
+                self,
+                #[upgrade_or_panic]
+                move |id| async move { self_.delete_handler(id).await }
+            ));
 
         let api = prefix
             .and(post_filter)
@@ -1066,7 +1042,7 @@ impl WhipServer {
             gst::debug!(CAT, "Stopped the server task...");
         });
 
-        gst::debug!(CAT, imp: self, "Started the server...");
+        gst::debug!(CAT, imp = self, "Started the server...");
         Some(jh)
     }
 
@@ -1079,7 +1055,7 @@ impl WhipServer {
 
 impl SignallableImpl for WhipServer {
     fn start(&self) {
-        gst::info!(CAT, imp: self, "starting the WHIP server");
+        gst::info!(CAT, imp = self, "starting the WHIP server");
         let jh = self.serve();
         let mut settings = self.settings.lock().unwrap();
         settings.server_handle = jh;
@@ -1099,20 +1075,25 @@ impl SignallableImpl for WhipServer {
             .expect("Shutdown signal Sender needs to be valid");
 
         if tx.send(()).is_err() {
-            gst::error!(CAT, imp: self, "Failed to send shutdown signal. Receiver dropped");
+            gst::error!(
+                CAT,
+                imp = self,
+                "Failed to send shutdown signal. Receiver dropped"
+            );
         }
 
-        gst::debug!(CAT, imp: self, "Await server handle to join");
+        gst::debug!(CAT, imp = self, "Await server handle to join");
         RUNTIME.block_on(async {
             if let Err(e) = handle.await {
-                gst::error!(CAT, imp:self, "Failed to join server handle: {e:?}");
+                gst::error!(CAT, imp = self, "Failed to join server handle: {e:?}");
             };
         });
 
-        gst::info!(CAT, imp: self, "stopped the WHIP server");
+        gst::info!(CAT, imp = self, "stopped the WHIP server");
     }
 
-    fn end_session(&self, _session_id: &str) {
+    fn end_session(&self, session_id: &str) {
+        gst::info!(CAT, imp = self, "Session {session_id} ended");
         //FIXME: send any events to the client
     }
 }
@@ -1127,26 +1108,27 @@ impl ObjectSubclass for WhipServer {
 
 impl ObjectImpl for WhipServer {
     fn properties() -> &'static [glib::ParamSpec] {
-        static PROPERTIES: Lazy<Vec<glib::ParamSpec>> = Lazy::new(|| {
+        static PROPERTIES: LazyLock<Vec<glib::ParamSpec>> = LazyLock::new(|| {
             vec![
+                glib::ParamSpecBoolean::builder("manual-sdp-munging")
+                    .nick("Manual SDP munging")
+                    .blurb("Whether the signaller manages SDP munging itself")
+                    .default_value(false)
+                    .read_only()
+                    .build(),
                 glib::ParamSpecString::builder("host-addr")
                     .nick("Host address")
                     .blurb("The the host address of the WHIP endpoint e.g., http://127.0.0.1:8080")
                     .default_value(DEFAULT_HOST_ADDR)
                     .flags(glib::ParamFlags::READWRITE)
                     .build(),
-                // needed by webrtcsrc in handle_webrtc_src_pad
-                glib::ParamSpecString::builder("producer-peer-id")
-                    .default_value(DEFAULT_PRODUCER_PEER_ID)
-                    .flags(glib::ParamFlags::READABLE)
-                    .build(),
                 glib::ParamSpecString::builder("stun-server")
                     .nick("STUN Server")
                     .blurb("The STUN server of the form stun://hostname:port")
                     .default_value(DEFAULT_STUN_SERVER)
                     .build(),
                 gst::ParamSpecArray::builder("turn-servers")
-                    .nick("List of TURN Servers to user")
+                    .nick("List of TURN Servers to use")
                     .blurb("The TURN servers of the form <\"turn(s)://username:password@host:port\", \"turn(s)://username1:password1@host1:port1\">")
                     .element_spec(&glib::ParamSpecString::builder("turn-server")
                         .nick("TURN Server")
@@ -1196,10 +1178,10 @@ impl ObjectImpl for WhipServer {
     fn property(&self, _id: usize, pspec: &glib::ParamSpec) -> glib::Value {
         let settings = self.settings.lock().unwrap();
         match pspec.name() {
+            "manual-sdp-munging" => false.to_value(),
             "host-addr" => settings.host_addr.to_string().to_value(),
             "stun-server" => settings.stun_server.to_value(),
             "turn-servers" => settings.turn_servers.to_value(),
-            "producer-peer-id" => settings.producer_peer_id.to_value(),
             "timeout" => settings.timeout.to_value(),
             _ => unimplemented!(),
         }
